; An M*CORE M210 emulator
; /Mic, 2005
;
; - Not all opcodes are implemented
; - Some opcodes have inaccurate timings (divs/divu/mult)
; - There may be endianness issues


.586p
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib

wsprintfA PROTO C :DWORD, :VARARG
wsprintf equ <wsprintfA>

.data

	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	CPU_REG				equ 00Ah
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h
	MMU_READ_DWORD_CALLBACK		equ 106h
	MMU_WRITE_DWORD_CALLBACK	equ 107h

	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h

	ZOMBIE_WINDOW_HANDLE		equ 500h

	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1


	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO

	cpu_inpb	PROTO :DWORD
	cpu_outpb	PROTO :DWORD,:DWORD

	
	Overflow	db 0
	Carry		db 0
	Zero		db 0
	Negative	db 0

	flags		db 0,0,0,0,0,0,0,1
	cycle		dd 0
	breakpoint	dd 0
	speed		dd 1400000h	; 20 MHz
	cpuName		db "M*CORE 210",0

	szBranch db "Branch instruction",13,10,0
	szTrue db "true",13,10,0
	szBuffer db 80 dup (0)
	szFormat db "Read %08x at pc=%08x",13,10,0
	szFormat2 db "Instruction type %d at %08x",13,10,0
	szFormat3 db "Writing %04x at %08x",13,10,0
	szFormat4 db "Branching with link from %08x to %08x",13,10,0
	szFormat5 db "mov pc,lr (%08x)",13,10,0
	szFormat9 db "MUL r%d,%08x,%08x",13,10,0
	szFormat10 db "%08x",13,10,0
	szFormat11 db "%08x, %08x, %08x, %08x, %08x, PC=%08x",13,10,0
	szFormat12 db "Offset = %08x",13,10,0
	szFormat13 db "PC=%08x, r0..r5 = %08x, %08x, %08x, %08x, %08x, %08x",13,10,0
	szFormat14 db "BT: Took branch after %d iterations (CMP %08x,%08x @ %08x)",13,10,0
	szFormat15 db "BF: Took branch after %d iterations (CMP %08x,%08x @ %08x)",13,10,0
	
.data?
	r0	dd ?	; stack pointer
	r1	dd ?	; 
	r2	dd ?	;
	r3	dd ?	; 
	r4	dd ?	; 
	r5	dd ?	; 
	r6	dd ?	; 
	r7	dd ?	; 
	r8	dd ?	; 
	r9	dd ?	; 
	r10	dd ?	; 
	r11	dd ?	; 
	r12	dd ?	; 
	r13	dd ?	; 	
	r14	dd ?	; 	
	r15	dd ?	; link/scratch	

	PC	dd ?	; program counter

	cond	dd ?
	PSR	dd ?
	VBR	dd ?
	EPSR	dd ?
	FPSR	dd ?
	EPC	dd ?
	FPC	dd ?
	SS0	dd ?
	SS1	dd ?
	SS2	dd ?
	SS3	dd ?
	SS4	dd ?
	GCR	dd ?
	GSR	dd ?
	
	bt_iter	dd ?
	bf_iter	dd ?
	
	hInst	dd ?
	hout	dd ?
	dummy 	dd ?
	lastPC 	dd ?
	lastCyc	dd ?
	physaddr dd ?
	physaddr_valid dd ?
	nextInstr dd ?
	nextAddr dd ?
	zombie	_4ARGS ?
	mmu_read_byte _1ARGS ?
	mmu_read_word _1ARGS ?
	mmu_read_dword _1ARGS ?
	mmu_write_byte _2ARGS ?
	mmu_write_word _2ARGS ?
	mmu_write_dword _2ARGS ?
	mmu_get _1ARGS ?
	mmu_translate_address _1ARGS ?
	
.code


ADD_CYCLES MACRO num
	add cycle,num ;*200h
ENDM


UPDATE_C MACRO
	setc byte ptr [cond]
ENDM

RETURN MACRO
	ret
ENDM

; Put Rd, Rs and Rt indexes in edx, ebx and ecx
GET_DST MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,21
	shr ecx,16
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM


DIV0 MACRO
	xor ecx,ecx
	div ecx
ENDM


;#####################################################################################################		

LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP


inpb_null PROC
	ret
inpb_null ENDP
.data
inpb_handlers dd 256 DUP(OFFSET inpb_null)
.code

outpb_null PROC dat:DWORD
	ret
outpb_null ENDP
.data
outpb_handlers dd 256 DUP(OFFSET outpb_null)
.code	


;#####################################################################################################		




op_nop:
	;ADD_CYCLES 1
	;RETURN
	DIV0

op_abs:
	and eax,0F00h
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	cmp ebx,0
	jne op_abs_pos
	neg ebx
	op_abs_pos:
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN

op_addc:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	add edx,cond
	add ebx,edx
	UPDATE_C
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_addi:
	mov ecx,eax
	mov edx,eax

	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	inc ecx
	bswap ebx
	add ebx,ecx
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_addu:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	add ebx,edx
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_and:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	and ebx,edx
	ADD_CYCLES 1
	mov [r0 + eax],ebx
	RETURN

op_andi:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	bswap ebx
	and ebx,ecx
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_andn:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	not edx

	and ebx,edx
	ADD_CYCLES 1
	mov [r0 + eax],ebx
	RETURN	


op_bf:
	.if cond == 0
	mov ebx,eax
	and ebx,0FF00h	; Low 8 bits
	and eax,7	; High 3 bits
	shl eax,29
	shl ebx,13
	or eax,ebx
	mov physaddr,-1
	sar eax,20
	add PC,eax
	ADD_CYCLES 2
	RETURN
	.endif
	ADD_CYCLES 1
	RETURN

op_br:
	mov ebx,eax
	and ebx,0FF00h	; Low 8 bits
	and eax,7	; High 3 bits
	shl eax,29
	shl ebx,13
	or eax,ebx
	mov physaddr,-1
	sar eax,20
	add PC,eax
	ADD_CYCLES 2
	RETURN

op_bsr:
	mov ebx,eax
	and ebx,0FF00h	; Low 8 bits
	and eax,7	; High 3 bits
	mov ecx,PC
	shl eax,29
	bswap ecx
	shl ebx,13
	mov r15,ecx	; Save PC in R15
	or eax,ebx
	mov physaddr,-1
	sar eax,20
	add PC,eax
	ADD_CYCLES 2
	RETURN
	
op_bt:
	.if cond == 1
	mov ebx,eax
	and ebx,0FF00h	; Low 8 bits
	and eax,7	; High 3 bits
	shl eax,29
	shl ebx,13
	or eax,ebx
	mov physaddr,-1
	sar eax,20
	add PC,eax
	ADD_CYCLES 2
	RETURN
	.endif
	ADD_CYCLES 1
	RETURN

op_clrf:
	.if cond == 0
	and eax,0F00h	; rX
	shr eax,6
	mov dword ptr [r0 + eax],0
	.endif
	ADD_CYCLES 1
	RETURN	

op_clrt:
	.if cond == 1
	and eax,0F00h	; rX
	shr eax,6
	mov dword ptr [r0 + eax],0
	.endif
	ADD_CYCLES 1
	RETURN	

op_cmphs:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	cmp ebx,edx
	setnc byte ptr [cond]
	ADD_CYCLES 1
	RETURN	

op_cmplt:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx

	bswap edx
	cmp ebx,edx
	setc byte ptr [cond]
	ADD_CYCLES 1
	RETURN	

op_cmpne:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	cmp ebx,edx
	setne byte ptr [cond]
	ADD_CYCLES 1
	RETURN	

op_cmpnei:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	bswap ebx
	cmp ebx,ecx
	setne byte ptr [cond]
	ADD_CYCLES 1
	RETURN

op_decf:
	.if cond == 0
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	sub ebx,1
	bswap ebx
	mov [r0 + eax],ebx
	.endif
	ADD_CYCLES 1
	RETURN	

op_decgt:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	sub ebx,1
	cmp ebx,0
	setg byte ptr [cond]
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN

op_dect:
	.if cond == 1
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	sub ebx,1
	bswap ebx
	mov [r0 + eax],ebx
	.endif
	ADD_CYCLES 1
	RETURN		
	

op_divs:
	mov ebx,eax
	and ebx,0F00h	; rX
	shr ebx,6
	mov eax,[r0 + ebx]
	mov ecx,r1
	bswap eax
	bswap ecx
	cdq
	idiv ecx
	bswap eax
	mov [r0 + ebx],eax
	ADD_CYCLES 35
	; TODO: Implement correct timing, division-by-zero exceptions
	RETURN	

op_divu:
	mov ebx,eax
	and ebx,0F00h	; rX
	shr ebx,6
	mov eax,[r0 + ebx]
	mov ecx,r1
	bswap eax
	bswap ecx
	cdq
	div ecx
	bswap eax
	mov [r0 + ebx],eax
	ADD_CYCLES 35
	; TODO: Implement correct timing, division-by-zero exceptions
	RETURN	
	
op_ff1:
	; TODO: Implement

op_incf:
	.if cond == 0
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	add ebx,1
	bswap ebx
	mov [r0 + eax],ebx
	.endif
	ADD_CYCLES 1
	RETURN	

op_inct:
	.if cond == 1
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	add ebx,1
	bswap ebx
	mov [r0 + eax],ebx
	.endif
	ADD_CYCLES 1
	RETURN	

op_ixh:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	lea ebx,[ebx + edx*2]
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_ixw:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	lea ebx,[ebx + edx*4]
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN	

op_jmp:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	and ebx,0FFFFFFFEh
	mov PC,ebx
	mov physaddr,-1
	ADD_CYCLES 2
	RETURN	


LDB_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,12
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_read_byte,ecx
	shl eax,24
	ADD_CYCLES 2
	mov rz,eax
	RETURN
ENDM

op_ldb_r0:
	LDB_REG r0
op_ldb_r1:
	LDB_REG r1
op_ldb_r2:
	LDB_REG r2
op_ldb_r3:
	LDB_REG r3
op_ldb_r4:
	LDB_REG r4
op_ldb_r5:
	LDB_REG r5
op_ldb_r6:
	LDB_REG r6
op_ldb_r7:
	LDB_REG r7
op_ldb_r8:
	LDB_REG r8
op_ldb_r9:
	LDB_REG r9
op_ldb_r10:
	LDB_REG r10
op_ldb_r11:
	LDB_REG r11
op_ldb_r12:
	LDB_REG r12
op_ldb_r13:
	LDB_REG r13
op_ldb_r14:
	LDB_REG r14
op_ldb_r15:
	LDB_REG r15
	

LDH_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,11
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_read_word,ecx
	shl eax,16
	ADD_CYCLES 2
	mov rz,eax
	RETURN
ENDM

op_ldh_r0:
	LDH_REG r0
op_ldh_r1:
	LDH_REG r1
op_ldh_r2:
	LDH_REG r2
op_ldh_r3:
	LDH_REG r3
op_ldh_r4:
	LDH_REG r4
op_ldh_r5:
	LDH_REG r5
op_ldh_r6:
	LDH_REG r6
op_ldh_r7:
	LDH_REG r7
op_ldh_r8:
	LDH_REG r8
op_ldh_r9:
	LDH_REG r9
op_ldh_r10:
	LDH_REG r10
op_ldh_r11:
	LDH_REG r11
op_ldh_r12:
	LDH_REG r12
op_ldh_r13:
	LDH_REG r13
op_ldh_r14:
	LDH_REG r14
op_ldh_r15:
	LDH_REG r15
	

LDW_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,10
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_read_dword,ecx
	mov rz,eax
	ADD_CYCLES 2
	RETURN
ENDM

op_ldw_r0:
	LDW_REG r0
op_ldw_r1:
	LDW_REG r1
op_ldw_r2:
	LDW_REG r2
op_ldw_r3:
	LDW_REG r3
op_ldw_r4:
	LDW_REG r4
op_ldw_r5:
	LDW_REG r5
op_ldw_r6:
	LDW_REG r6
op_ldw_r7:
	LDW_REG r7
op_ldw_r8:
	LDW_REG r8
op_ldw_r9:
	LDW_REG r9
op_ldw_r10:
	LDW_REG r10
op_ldw_r11:
	LDW_REG r11
op_ldw_r12:
	LDW_REG r12
op_ldw_r13:
	LDW_REG r13
op_ldw_r14:
	LDW_REG r14
op_ldw_r15:
	LDW_REG r15
	
	
LRW_REG MACRO rz
	and eax,0FF00h
	shr eax,6
	add eax,PC
	and eax,0FFFFFFFCh
	invoke mmu_read_dword,eax
	mov rz,eax
	ADD_CYCLES 2
	RETURN
ENDM

op_lrw_r0:
	LRW_REG r0
op_lrw_r1:
	LRW_REG r1
op_lrw_r2:
	LRW_REG r2
op_lrw_r3:
	LRW_REG r3
op_lrw_r4:
	LRW_REG r4
op_lrw_r5:
	LRW_REG r5
op_lrw_r6:
	LRW_REG r6
op_lrw_r7:
	LRW_REG r7
op_lrw_r8:
	LRW_REG r8
op_lrw_r9:
	LRW_REG r9
op_lrw_r10:
	LRW_REG r10
op_lrw_r11:
	LRW_REG r11
op_lrw_r12:
	LRW_REG r12
op_lrw_r13:
	LRW_REG r13
op_lrw_r14:
	LRW_REG r14
op_lrw_r15:
	LRW_REG r15

op_lslc:
	ADD_CYCLES 1
	shl ebx,1
	setc byte ptr [cond]
	bswap ebx
	mov [r0 + eax],ebx
	RETURN	
op_lsli:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	bswap ebx
	test ecx,ecx
	jz op_lslc
	shl ebx,cl
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_lsrc:
	ADD_CYCLES 1
	shr ebx,1
	setc byte ptr [cond]
	bswap ebx
	mov [r0 + eax],ebx
	RETURN	
op_lsri:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	bswap ebx
	test ecx,ecx
	jz op_lsrc
	shr ebx,cl
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_mov:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov edx,[r0 + ecx]
	mov [r0 + eax],edx
	ADD_CYCLES 1
	RETURN
	
op_movf:
	.if cond == 0
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov edx,[r0 + ecx]
	mov [r0 + eax],edx
	.endif
	ADD_CYCLES 1
	RETURN

op_movi:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm7
	and edx,7
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	bswap ecx

	mov [r0 + eax],ecx
	ADD_CYCLES 1
	RETURN

op_movt:
	.if cond == 1
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov edx,[r0 + ecx]
	mov [r0 + eax],edx
	.endif
	ADD_CYCLES 1
	RETURN

op_mult:
	mov ecx,eax
	mov ebx,eax
	and ecx,0F00h	; rX
	and ebx,0F000h	; rY
	shr ecx,6
	shr ebx,10
	mov eax,[r0 + ebx]
	mov ebx,[r0 + ecx]
	bswap eax
	bswap ebx
	mul ebx
	bswap eax
	mov [r0 + ecx],eax
	; TODO: Implement correct timing
	ADD_CYCLES 18
	RETURN

op_or:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	or ebx,[r0 + ecx]
	ADD_CYCLES 1
	mov [r0 + eax],ebx
	RETURN	

op_rsubi:
	mov ecx,eax
	mov edx,eax
	and ecx,0F000h	; imm5
	and edx,1
	and eax,0F00h	; rX
	shl edx,4
	shr eax,6
	shr ecx,12
	or ecx,edx
	mov ebx,[r0 + eax]
	bswap ebx
	sub ecx,ebx
	ADD_CYCLES 1
	bswap ecx
	mov [r0 + eax],ecx
	RETURN

op_sextb:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	movsx ebx,bl
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN	

op_sexth:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	movsx ebx,bx
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN	


STB_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,12
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_write_byte,ecx,dword ptr [rz + 3]
	ADD_CYCLES 2
	RETURN
ENDM

op_stb_r0:
	STB_REG r0
op_stb_r1:
	STB_REG r1
op_stb_r2:
	STB_REG r2
op_stb_r3:
	STB_REG r3
op_stb_r4:
	STB_REG r4
op_stb_r5:
	STB_REG r5
op_stb_r6:
	STB_REG r6
op_stb_r7:
	STB_REG r7
op_stb_r8:
	STB_REG r8
op_stb_r9:
	STB_REG r9
op_stb_r10:
	STB_REG r10
op_stb_r11:
	STB_REG r11
op_stb_r12:
	STB_REG r12
op_stb_r13:
	STB_REG r13
op_stb_r14:
	STB_REG r14
op_stb_r15:
	STB_REG r15
	
	
STW_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,10
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_write_dword,ecx,dword ptr [rz]
	ADD_CYCLES 2
	RETURN
ENDM

op_stw_r0:
	STW_REG r0
op_stw_r1:
	STW_REG r1
op_stw_r2:
	STW_REG r2
op_stw_r3:
	STW_REG r3
op_stw_r4:
	STW_REG r4
op_stw_r5:
	STW_REG r5
op_stw_r6:
	STW_REG r6
op_stw_r7:
	STW_REG r7
op_stw_r8:
	STW_REG r8
op_stw_r9:
	STW_REG r9
op_stw_r10:
	STW_REG r10
op_stw_r11:
	STW_REG r11
op_stw_r12:
	STW_REG r12
op_stw_r13:
	STW_REG r13
op_stw_r14:
	STW_REG r14
op_stw_r15:
	STW_REG r15


STH_REG MACRO rz
	mov ebx,eax
	and ebx,00F00h	; rX
	and eax,0F000h	; imm4
	shr ebx,6
	shr eax,11
	mov ecx,[r0 + ebx]
	bswap ecx
	add ecx,eax
	invoke mmu_write_word,ecx,dword ptr [rz + 2]
	ADD_CYCLES 2
	RETURN
ENDM

op_sth_r0:
	STH_REG r0
op_sth_r1:
	STH_REG r1
op_sth_r2:
	STH_REG r2
op_sth_r3:
	STH_REG r3
op_sth_r4:
	STH_REG r4
op_sth_r5:
	STH_REG r5
op_sth_r6:
	STH_REG r6
op_sth_r7:
	STH_REG r7
op_sth_r8:
	STH_REG r8
op_sth_r9:
	STH_REG r9
op_sth_r10:
	STH_REG r10
op_sth_r11:
	STH_REG r11
op_sth_r12:
	STH_REG r12
op_sth_r13:
	STH_REG r13
op_sth_r14:
	STH_REG r14
op_sth_r15:
	STH_REG r15
	

op_subu:
	mov ecx,eax
	and eax,0F00h	; rX
	and ecx,0F000h	; rY
	shr eax,6
	shr ecx,10
	mov ebx,[r0 + eax]
	mov edx,[r0 + ecx]
	bswap ebx
	bswap edx
	sub ebx,edx
	ADD_CYCLES 1
	bswap ebx
	mov [r0 + eax],ebx
	RETURN

op_zextb:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	movzx ebx,bl
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN	

op_zexth:
	and eax,0F00h	; rX
	shr eax,6
	mov ebx,[r0 + eax]
	bswap ebx
	movzx ebx,bx
	bswap ebx
	mov [r0 + eax],ebx
	ADD_CYCLES 1
	RETURN	
	
	
.data
align 8
op_00_table LABEL DWORD
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_dect
	dd OFFSET op_decf
	dd OFFSET op_inct
	dd OFFSET op_incf
	dd OFFSET op_jmp
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
.code
op_00:
	mov ebx,eax
	and ebx,0F000h
	shr ebx,10
	jmp dword ptr [op_00_table + ebx]
	
.data
align 8
op_01_table LABEL DWORD
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_zextb
	dd OFFSET op_sextb
	dd OFFSET op_zexth
	dd OFFSET op_sexth
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_decgt
	dd OFFSET op_nop
	dd OFFSET op_clrt
	dd OFFSET op_clrf
	dd OFFSET op_nop
	dd OFFSET op_nop
.code
op_01:
	mov ebx,eax
	and ebx,0F000h
	shr ebx,10
	jmp dword ptr [op_01_table + ebx]

.data
align 8
op_2C_table LABEL DWORD
	dd OFFSET op_nop
	dd OFFSET op_divu
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
.code
op_2C:
	mov ebx,eax
	and ebx,0F000h
	shr ebx,10
	jmp dword ptr [op_2C_table + ebx]	

.data
align 8
op_32_table LABEL DWORD
	dd OFFSET op_nop
	dd OFFSET op_divs
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
	dd OFFSET op_nop
.code
op_32:
	mov ebx,eax
	and ebx,0F000h
	shr ebx,10
	jmp dword ptr [op_32_table + ebx]	
	
;###########################################################################################

.data
align 16
op_table LABEL DWORD
; 00-0F
dd OFFSET op_00
dd OFFSET op_01
dd OFFSET op_movt
dd OFFSET op_mult
dd OFFSET op_nop
dd OFFSET op_subu
dd OFFSET op_addc
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_movf
dd OFFSET op_nop
dd OFFSET op_cmphs
dd OFFSET op_cmplt
dd OFFSET op_nop
dd OFFSET op_cmpne

; 10-1F
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_mov
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_ixw
dd OFFSET op_and
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_addu
dd OFFSET op_ixh
dd OFFSET op_or
dd OFFSET op_andn

; 20-2F
dd OFFSET op_addi
dd OFFSET op_addi
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_rsubi
dd OFFSET op_rsubi
dd OFFSET op_cmpnei
dd OFFSET op_cmpnei
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_andi
dd OFFSET op_andi

; 30-3F
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_lsli
dd OFFSET op_lsli
dd OFFSET op_lsri
dd OFFSET op_lsri

; 40-4F
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop

; 50-5F
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop

; 60-6F
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_movi
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop
dd OFFSET op_nop

; 70-7F
dd OFFSET op_lrw_r0
dd OFFSET op_lrw_r1
dd OFFSET op_lrw_r2
dd OFFSET op_lrw_r3
dd OFFSET op_lrw_r4
dd OFFSET op_lrw_r5
dd OFFSET op_lrw_r6
dd OFFSET op_lrw_r7
dd OFFSET op_lrw_r8
dd OFFSET op_lrw_r9
dd OFFSET op_lrw_r10
dd OFFSET op_lrw_r11
dd OFFSET op_lrw_r12
dd OFFSET op_lrw_r13
dd OFFSET op_lrw_r14
dd OFFSET op_lrw_r15

; 80-8F
dd OFFSET op_ldw_r0
dd OFFSET op_ldw_r1
dd OFFSET op_ldw_r2
dd OFFSET op_ldw_r3
dd OFFSET op_ldw_r4
dd OFFSET op_ldw_r5
dd OFFSET op_ldw_r6
dd OFFSET op_ldw_r7
dd OFFSET op_ldw_r8
dd OFFSET op_ldw_r9
dd OFFSET op_ldw_r10
dd OFFSET op_ldw_r11
dd OFFSET op_ldw_r12
dd OFFSET op_ldw_r13
dd OFFSET op_ldw_r14
dd OFFSET op_ldw_r15

; 90-9F
dd OFFSET op_stw_r0
dd OFFSET op_stw_r1
dd OFFSET op_stw_r2
dd OFFSET op_stw_r3
dd OFFSET op_stw_r4
dd OFFSET op_stw_r5
dd OFFSET op_stw_r6
dd OFFSET op_stw_r7
dd OFFSET op_stw_r8
dd OFFSET op_stw_r9
dd OFFSET op_stw_r10
dd OFFSET op_stw_r11
dd OFFSET op_stw_r12
dd OFFSET op_stw_r13
dd OFFSET op_stw_r14
dd OFFSET op_stw_r15

; A0-AF
dd OFFSET op_ldb_r0
dd OFFSET op_ldb_r1
dd OFFSET op_ldb_r2
dd OFFSET op_ldb_r3
dd OFFSET op_ldb_r4
dd OFFSET op_ldb_r5
dd OFFSET op_ldb_r6
dd OFFSET op_ldb_r7
dd OFFSET op_ldb_r8
dd OFFSET op_ldb_r9
dd OFFSET op_ldb_r10
dd OFFSET op_ldb_r11
dd OFFSET op_ldb_r12
dd OFFSET op_ldb_r13
dd OFFSET op_ldb_r14
dd OFFSET op_ldb_r15

; B0-BF
dd OFFSET op_stb_r0
dd OFFSET op_stb_r1
dd OFFSET op_stb_r2
dd OFFSET op_stb_r3
dd OFFSET op_stb_r4
dd OFFSET op_stb_r5
dd OFFSET op_stb_r6
dd OFFSET op_stb_r7
dd OFFSET op_stb_r8
dd OFFSET op_stb_r9
dd OFFSET op_stb_r10
dd OFFSET op_stb_r11
dd OFFSET op_stb_r12
dd OFFSET op_stb_r13
dd OFFSET op_stb_r14
dd OFFSET op_stb_r15

; C0-CF
dd OFFSET op_ldh_r0
dd OFFSET op_ldh_r1
dd OFFSET op_ldh_r2
dd OFFSET op_ldh_r3
dd OFFSET op_ldh_r4
dd OFFSET op_ldh_r5
dd OFFSET op_ldh_r6
dd OFFSET op_ldh_r7
dd OFFSET op_ldh_r8
dd OFFSET op_ldh_r9
dd OFFSET op_ldh_r10
dd OFFSET op_ldh_r11
dd OFFSET op_ldh_r12
dd OFFSET op_ldh_r13
dd OFFSET op_ldh_r14
dd OFFSET op_ldh_r15

; D0-DF
dd OFFSET op_sth_r0
dd OFFSET op_sth_r1
dd OFFSET op_sth_r2
dd OFFSET op_sth_r3
dd OFFSET op_sth_r4
dd OFFSET op_sth_r5
dd OFFSET op_sth_r6
dd OFFSET op_sth_r7
dd OFFSET op_sth_r8
dd OFFSET op_sth_r9
dd OFFSET op_sth_r10
dd OFFSET op_sth_r11
dd OFFSET op_sth_r12
dd OFFSET op_sth_r13
dd OFFSET op_sth_r14
dd OFFSET op_sth_r15

; E0-EF
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bt
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf
dd OFFSET op_bf

; F0-FF
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_br
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr
dd OFFSET op_bsr




.code


cpu_init PROC commlink:DWORD	
	mov eax,commlink
	mov zombie,eax

	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax

	pusha	
	mov edi,offset r0
	mov ecx,16
	mov eax,0
	rep stosd
	popa
	
	ret
cpu_init ENDP



cpu_reset PROC
	pusha

	invoke zombie,ASK,MMU_READ_BYTE_CALLBACK,0,0
	mov mmu_read_byte,eax
	invoke zombie,ASK,MMU_READ_WORD_CALLBACK,0,0
	mov mmu_read_word,eax
	invoke zombie,ASK,MMU_READ_DWORD_CALLBACK,0,0
	mov mmu_read_dword,eax

	invoke zombie,ASK,MMU_WRITE_BYTE_CALLBACK,0,0
	mov mmu_write_byte,eax
	invoke zombie,ASK,MMU_WRITE_WORD_CALLBACK,0,0
	mov mmu_write_word,eax
	invoke zombie,ASK,MMU_WRITE_DWORD_CALLBACK,0,0
	mov mmu_write_dword,eax

	invoke zombie,ASK,MMU_TRANSLATE_ADDRESS_CALLBACK,0,0
	mov mmu_translate_address,eax

	mov edi,offset r0
	mov ecx,16
	mov eax,0
	rep stosd


	mov physaddr_valid,0
	
	mov bf_iter,0
	mov bt_iter,0
	
	popa
	ret
cpu_reset ENDP


cpu_close PROC
	ret
cpu_close ENDP


cpu_config PROC hWnd:DWORD
	ret
cpu_config ENDP


cpu_set_io_handler PROC mode:DWORD,port:DWORD,pfhandler:DWORD
	push ebx

	mov eax,port
	mov ebx,pfhandler
	
	.if mode==INPUT	
		mov [inpb_handlers + eax*4],ebx
	.else
		mov [outpb_handlers + eax*4],ebx
	.endif
	
	pop ebx
	ret
cpu_set_io_handler ENDP


cpu_get PROC what:DWORD
	.if what == CPU_CYCLE_COUNT
		mov eax,cycle
	.elseif what == CPU_CYCLE_SCALING
		mov eax,1 ;200h
	.elseif what == CPU_SET_IO_CALLBACK
		mov eax,offset cpu_set_io_handler
	.elseif what == CPU_IRQ_CALLBACK
		mov eax,offset cpu_irq
	.elseif what == CPU_INPB_CALLBACK
		mov eax,offset cpu_inpb
	.elseif what == CPU_OUTPB_CALLBACK
		mov eax,offset cpu_outpb
	.elseif what == CPU_SPEED
		mov eax,speed
	.elseif what == CPU_REG_PC
		mov eax,PC
	.elseif what == CPU_REG
		mov eax,OFFSET r0
	.elseif what == PLUGIN_NAME
		mov eax,offset cpuName
	

	.elseif what == 020h
		pusha
		invoke wsprintf,ADDR szBuffer,ADDR szFormat13,PC,r0,r1,r2,r3,r4,r5
		invoke lstrlen,ADDR szBuffer
		invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
		popa
	.endif
	ret
cpu_get ENDP


cpu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == CPU_REG_PC
		mov eax,param1
		mov PC,eax
	.elseif what == CPU_REG_SP
		mov eax,param1
		mov r0,eax
	.elseif what == CPU_CYCLE_COUNT
		mov eax,param1
		mov cycle,eax
	.elseif what == CPU_BREAKPOINT
		mov eax,param1
		mov breakpoint,eax
	.elseif what == CPU_SPEED
		mov eax,param1
		mov speed,eax
	.elseif what == CPU_IO_HANDLER
		mov eax,param1
		and param1,0FFh
		shr eax,8
		invoke cpu_set_io_handler,eax,param1,param2
	.endif
	ret
cpu_set ENDP



cpu_outpb PROC port:DWORD,dat:DWORD
	push ecx
	push dword ptr dat
	mov eax,port
	call dword ptr [outpb_handlers + eax*4]
	pop ecx
	ret
cpu_outpb ENDP


cpu_inpb PROC port:DWORD
	push ecx
	mov eax,port
	call dword ptr [inpb_handlers + eax*4]
	pop ecx
	ret
cpu_inpb ENDP


cpu_irq PROC address:DWORD
	ret
cpu_irq ENDP


cpu_execute PROC
	ret
cpu_execute ENDP


align 8

; In:  cycle threshold
; Out: executed cycles
PUBLIC cpu_execute_until
cpu_execute_until: 		
	
	pusha
	
	mov ecx,cycle			; Put cycle counter in ecx
	mov physaddr,-1			; Invalidate physical address mapping
	
	@@exe_loop:
		cmp ecx,[esp+36]	; Is the cycle count > threshold ?
		jge @@exe_done
		;mov lastCyc,ecx

		;pusha
		;invoke wsprintf,ADDR szBuffer,ADDR szFormat13,PC,r0,r1,r2,r3,r4,r5
		;invoke lstrlen,ADDR szBuffer
		;invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
		;popa
		
		mov eax,physaddr
		cmp eax,-1
		jne @@physaddrvalid
			mov eax,PC
			invoke mmu_translate_address,eax
			mov physaddr,eax
		@@physaddrvalid:
		movzx ebx,word ptr [eax]
		add physaddr,2
		add PC,2

		; Instruction format:
		;	xxxx nnnn mmmm iidd  (big-endian (sh-2))
		;	mmmm iidd xxxx nnnn  (little-endian (x86))
		mov eax,ebx
		and ebx,0FFh
		call dword ptr [op_table + ebx*4]
		
		@@next_instr::
		
		;sub PC,2
		
		mov ecx,cycle
		jmp @@exe_loop
	@@exe_done:
	popa
	
	mov eax,cycle
	ret 4



END LibMain
