; An MIPS R2000 emulator
; /Mic, 2004


.586p
.model flat,stdcall
option casemap:none


include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
include \masm32\include\user32.inc

includelib \masm32\lib\kernel32.lib
includelib \masm32\lib\user32.lib

wsprintfA PROTO C :DWORD, :VARARG
wsprintf equ <wsprintfA>

.data

	CPU_IRQ_CALLBACK		equ 000h
	CPU_SET_IO_CALLBACK		equ 001h
	CPU_IO_HANDLER			equ 001h
	CPU_CYCLE_COUNT			equ 002h
	CPU_REG_PC			equ 003h
	CPU_REG_SP			equ 004h
	CPU_BREAKPOINT			equ 005h
	CPU_SPEED			equ 006h
	CPU_INPB_CALLBACK		equ 007h
	CPU_OUTPB_CALLBACK		equ 008h
	CPU_CYCLE_SCALING		equ 009h
	CPU_REG				equ 00Ah
	
	MMU_READ_BYTE_CALLBACK		equ 100h
	MMU_READ_WORD_CALLBACK		equ 101h
	MMU_WRITE_BYTE_CALLBACK		equ 102h
	MMU_WRITE_WORD_CALLBACK		equ 103h
	MMU_TRANSLATE_ADDRESS_CALLBACK	equ 104h
	MMU_READ_DWORD_CALLBACK		equ 106h
	MMU_WRITE_DWORD_CALLBACK	equ 107h

	GPU_CURRENT_SCANLINE		equ 200h
	GPU_SCREEN_WIDTH		equ 201h
	GPU_SCREEN_HEIGHT		equ 202h
	GPU_FRAME_COMPLETED		equ 203h
	GPU_SCREEN_PITCH		equ 204h

	ZOMBIE_WINDOW_HANDLE		equ 500h

	PLUGIN_NAME			equ 600h
	
	INPUT		equ 0
	OUTPUT		equ 1

	ASK		equ 0
	TELL		equ 1


	_1ARGS_PROTO	typedef PROTO :DWORD
	_2ARGS_PROTO	typedef PROTO :DWORD,:DWORD
	_4ARGS_PROTO	typedef PROTO :DWORD,:DWORD,:DWORD,:DWORD
	_1ARGS		typedef ptr _1ARGS_PROTO
	_2ARGS		typedef ptr _2ARGS_PROTO
	_4ARGS		typedef ptr _4ARGS_PROTO

	cpu_inpb	PROTO :DWORD
	cpu_outpb	PROTO :DWORD,:DWORD

	
	Overflow	db 0
	Carry		db 0
	Zero		db 0
	Negative	db 0

	flags		db 0,0,0,0,0,0,0,1
	cycle		dd 0
	breakpoint	dd 0
	speed		dd 1400000h	; 20 MHz
	cpuName		db "MIPS R2000",0

	szBranch db "Branch instruction",13,10,0
	szTrue db "true",13,10,0
	szBuffer db 80 dup (0)
	szFormat db "Read %08x at pc=%08x",13,10,0
	szFormat2 db "Instruction type %d at %08x",13,10,0
	szFormat3 db "Writing %04x at %08x",13,10,0
	szFormat4 db "Branching with link from %08x to %08x",13,10,0
	szFormat5 db "mov pc,lr (%08x)",13,10,0
	szFormat6 db "stmfd %08x",13,10,0
	szFormat7 db "ldmfd %08x",13,10,0
	szFormat8 db "ldr r%d,=%d",13,10,0
	szFormat9 db "MUL r%d,%08x,%08x",13,10,0
	szFormat10 db "%08x",13,10,0
	szFormat11 db "%08x, %08x, %08x, %08x, %08x, PC=%08x",13,10,0
	
	
	
.data?
	r0	dd ?	; r0
	r1	dd ?	; at
	r2	dd ?	; v0
	r3	dd ?	; v1
	r4	dd ?	; a0
	r5	dd ?	; a1
	r6	dd ?	; a2
	r7	dd ?	; a3
	r8	dd ?	; t0
	r9	dd ?	; t1
	r10	dd ?	; t2
	r11	dd ?	; t3
	r12	dd ?	; t4
	r13	dd ?	; t5	
	r14	dd ?	; t6	
	r15	dd ?	; t7	
	r16	dd ?	; s0
	r17	dd ?	; s1
	r18	dd ?	; s2
	r19	dd ?	; s3
	r20	dd ?	; s4
	r21	dd ?	; s5
	r22	dd ?	; s6
	r23	dd ?	; s7
	r24	dd ?	; t8
	r25	dd ?	; t9
	r26	dd ?	; k0
	r27	dd ?	; k1
	r28	dd ?	; gp
	r29	dd ?	; sp
	r30	dd ?	; s8	
	r31	dd ?	; ra	

	_hi	dd ?
	_lo	dd ?
	
	PC	dd ?
	EPC	dd ?
	Cause	dd ?
	Status	dd ?

	UNKNOWN_INSTRUCTION EQU 0
	ARITHMETIC_OVERFLOW EQU 1
	
	f0	dd ?
	f1	dd ?
	f2	dd ?
	f3	dd ?
	f4	dd ?
	f5	dd ?
	f6	dd ?
	f7	dd ?
	f8	dd ?
	f9	dd ?
	f10	dd ?
	f11	dd ?
	f12	dd ?
	f13	dd ?
	f14	dd ?
	f15	dd ?
	f16	dd ?
	f17	dd ?
	f18	dd ?
	f19	dd ?
	f20	dd ?
	f21	dd ?
	f22	dd ?
	f23	dd ?
	f24	dd ?
	f25	dd ?
	f26	dd ?
	f27	dd ?
	f28	dd ?
	f29	dd ?
	f30	dd ?
	f31	dd ?
	
	fflag	dd ?
	
	hInst	dd ?
	hout	dd ?
	dummy 	dd ?
	lastPC 	dd ?
	lastCyc	dd ?
	physaddr dd ?
	nextInstr dd ?
	nextAddr dd ?
	zombie	_4ARGS ?
	mmu_read_byte _1ARGS ?
	mmu_read_word _1ARGS ?
	mmu_read_dword _1ARGS ?
	mmu_write_byte _2ARGS ?
	mmu_write_word _2ARGS ?
	mmu_write_dword _2ARGS ?
	mmu_get _1ARGS ?
	mmu_translate_address _1ARGS ?
	
.code


ADD_CYCLES MACRO num
	add cycle,num*200h
ENDM

	

COND_BRANCH MACRO jump
	LOCAL cond_true

	mov physaddr,-1
	
	GET_ST
	mov ebx,[r0+ebx*4]
	cmp ebx,[r0+ecx*4]
	jump cond_true
	ADD_CYCLES 1
	RETURN
	
	cond_true:
	shl eax,2
	and eax,0FFFCh
	movsx eax,ax
	add PC,eax
	ADD_CYCLES 2
	EXECUTE_DBSI
	RETURN
ENDM


COND_BRANCH_ZR MACRO jump
	LOCAL cond_true

	mov physaddr,-1
	
	GET_S
	mov ebx,[r0+ebx*4]
	cmp ebx,0
	jump cond_true
	ADD_CYCLES 1
	RETURN
	
	cond_true:
	shl eax,2
	and eax,0FFFCh
	movsx eax,ax
	add PC,eax
	ADD_CYCLES 2
	EXECUTE_DBSI
	RETURN
ENDM


; Put Rd, Rs and Rt indexes in edx, ebx and ecx
GET_DST MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,21
	shr ecx,16
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM

; Put Rd, Rs and Sa indexes in edx, ebx and ecx
GET_DSA MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,21
	shr ecx,6
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM


; Put Rd, Rt and Sa indexes in edx, ebx and ecx
GET_DTA MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,16
	shr ecx,6
	shr edx,11
	and ebx,31
	and ecx,31
	and edx,31
ENDM


GET_ST MACRO
	mov ebx,eax
	mov ecx,eax
	shr ebx,21
	shr ecx,16
	and ebx,31
	and ecx,31
ENDM

GET_S MACRO
	mov ebx,eax
	shr ebx,21
	and ebx,31
ENDM

GET_T MACRO
	mov ecx,eax
	shr ecx,16
	and ecx,31
ENDM

GET_D MACRO
	mov edx,eax
	shr edx,11
	and edx,31
ENDM


; Put Rd, Rs and Rt indexes in edx, ebx and ecx
GET_COP_DST MACRO
	mov ebx,eax
	mov ecx,eax
	mov edx,eax
	shr ebx,11
	shr ecx,16
	shr edx,6
	and ebx,31
	and ecx,31
	and edx,31
ENDM

GET_COP_ST MACRO
	mov ebx,eax
	mov ecx,eax
	shr ebx,11
	shr ecx,16
	and ebx,31
	and ecx,31
ENDM

RETURN MACRO
;	ret
	jmp ebp
ENDM

; Execute the instruction in the delayed branch slot
EXECUTE_DBSI MACRO
	LOCAL not_rtype

	;push ebp
	;mov eax,lastPC
	;invoke mmu_read_dword,eax
	;pop ebp	
	mov eax,nextInstr
	
	.if eax != 0		; Skip NOPs
		test eax,0FC000000h
		jnz not_rtype
		mov ebx,eax
			and ebx,03Fh
			jmp dword ptr [rtype_op + ebx*4]
		not_rtype:
			mov ebx,eax
			shr ebx,26
			and ebx,03Fh
			jmp dword ptr [opcode + ebx*4]
	.else
		ADD_CYCLES 1
	.endif
ENDM

		

LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD

        .if reason == 1		;DLL_PROCESS_ATTACH
        	mov eax,hInstDLL
        	mov hInst,eax
		mov eax,TRUE
		ret

        .elseif reason == 0	;DLL_PROCESS_DETACH
        .elseif reason == 2	;DLL_THREAD_ATTACH
        .elseif reason == 3	;DLL_THREAD_DETACH
        .endif

        mov eax,FALSE
        ret
LibMain ENDP


inpb_null PROC
	ret
inpb_null ENDP
.data
inpb_handlers dd 256 DUP(OFFSET inpb_null)
.code

outpb_null PROC dat:DWORD
	ret
outpb_null ENDP
.data
outpb_handlers dd 256 DUP(OFFSET outpb_null)
.code	



op_and:
	GET_DST
	mov ebx,[r0+ebx*4]
	and ebx,[r0+ecx*4]
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_add:
	GET_DST
	mov ebx,[r0+ebx*4]
	add ebx,[r0+ecx*4]
	jo @@op_add_overflow
		mov [r0+edx*4],ebx
	@@op_add_overflow:
	ADD_CYCLES 1
	RETURN
	
op_addu:
	GET_DST
	mov ebx,[r0+ebx*4]
	add ebx,[r0+ecx*4]
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_break:
op_div:
	GET_ST
	mov eax,[r0+ebx*4]
	cdq
	idiv dword ptr [r0+ecx*4]
	jo @@op_div_overflow
		mov _hi,edx
		mov _lo,eax
	@@op_div_overflow:
	ADD_CYCLES 35
	RETURN
	
op_divu:
	GET_ST
	mov eax,[r0+ebx*4]
	cdq
	idiv dword ptr [r0+ecx*4]
	mov _hi,edx
	mov _lo,eax
	ADD_CYCLES 35
	RETURN

op_jalr:

op_jr:
	GET_S
	mov eax,[r0+ebx*4]
	mov physaddr,-1
	mov PC,eax
	ADD_CYCLES 2
	EXECUTE_DBSI
	RETURN
	
op_mfhi:
	GET_D
	mov ebx,_hi
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_mflo:
	GET_D
	mov ebx,_lo
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_mthi:
op_mtlo:

op_mult:
	GET_ST
	mov eax,[r0+ebx*4]
	imul dword ptr [r0+ecx*4]
	jo @@op_mult_overflow
		mov _hi,edx
		mov _lo,eax
	@@op_mult_overflow:
	ADD_CYCLES 2
	RETURN
	
op_multu:
	GET_ST
	mov eax,[r0+ebx*4]
	imul dword ptr [r0+ecx*4]
	mov _hi,edx
	mov _lo,eax
	ADD_CYCLES 2
	RETURN
	
op_nor:
	GET_DST
	mov ebx,[r0+ebx*4]
	or ebx,[r0+ecx*4]
	not ebx

	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_or:
	GET_DST
	mov ebx,[r0+ebx*4]
	or ebx,[r0+ecx*4]
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_sll:
	GET_DTA
	mov ebx,[r0+ebx*4]
	shl ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_sllv:
	GET_DST
	mov ebx,[r0+ebx*4]
	mov ecx,[r0+ecx*4]
	shl ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_slt:
	GET_DST
	mov eax,[r0+ebx*4]
	cmp eax,[r0+ecx*4]
	setl bl
	and ebx,1
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
	
op_sltu:
	GET_DST
	mov eax,[r0+ebx*4]
	cmp eax,[r0+ecx*4]
	setb bl
	and ebx,1
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_sra:
	GET_DTA
	mov ebx,[r0+ebx*4]
	sar ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_srav:
	GET_DST
	mov ebx,[r0+ebx*4]
	mov ecx,[r0+ecx*4]
	sar ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_srl:
	GET_DTA
	mov ebx,[r0+ebx*4]
	shr ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
op_srlv:
	GET_DST
	mov ebx,[r0+ebx*4]
	mov ecx,[r0+ecx*4]
	shr ebx,cl
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN
	
op_sub:
	GET_DST
	mov ebx,[r0+ebx*4]
	sub ebx,[r0+ecx*4]
	jo @@op_sub_overflow
		mov [r0+edx*4],ebx
	@@op_sub_overflow:
	ADD_CYCLES 1
	RETURN

op_subu:
	GET_DST
	mov ebx,[r0+ebx*4]
	sub ebx,[r0+ecx*4]
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_syscall:
	.if r2 == 1
		pusha
		invoke wsprintf,ADDR szBuffer,ADDR szFormat10,r4
		invoke lstrlen,ADDR szBuffer
		invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
		popa
	.endif
	ADD_CYCLES 1
	RETURN
	
op_xor:
	GET_DST
	mov ebx,[r0+ebx*4]
	xor ebx,[r0+ecx*4]
	mov [r0+edx*4],ebx
	ADD_CYCLES 1
	RETURN

op_unknown:
	mov eax,PC
	sub eax,4
	mov EPC,eax
	mov Cause,UNKNOWN_INSTRUCTION
	RETURN



rtype_op:
	dd OFFSET op_sll
	dd OFFSET op_unknown	
	dd OFFSET op_srl
	dd OFFSET op_sra
	dd OFFSET op_sllv
	dd OFFSET op_unknown
	dd OFFSET op_srlv
	dd OFFSET op_srav
	dd OFFSET op_jr
	dd OFFSET op_jalr
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_syscall
	dd OFFSET op_break
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	
	dd OFFSET op_mfhi
	dd OFFSET op_mthi
	dd OFFSET op_mflo
	dd OFFSET op_mtlo
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_mult
	dd OFFSET op_multu
	dd OFFSET op_div
	dd OFFSET op_divu
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	

	dd OFFSET op_add
	dd OFFSET op_addu
	dd OFFSET op_sub
	dd OFFSET op_subu
	dd OFFSET op_and
	dd OFFSET op_or
	dd OFFSET op_xor
	dd OFFSET op_nor
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_slt
	dd OFFSET op_sltu
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	

	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown
	dd OFFSET op_unknown


opcode_00:

opcode_01:
	test eax,10000h
	jnz @@bgez
		COND_BRANCH_ZR jl
	@@bgez:
		COND_BRANCH_ZR jge
	
; J
opcode_02:
	and eax,003FFFFFFh
	mov ebx,PC
	shl eax,2
	and ebx,0F0000000h
	mov physaddr,-1
	or eax,ebx
	mov PC,eax
	ADD_CYCLES 2
	EXECUTE_DBSI
	RETURN

; JAL
opcode_03:
	and eax,003FFFFFFh
	mov ebx,PC
	shl eax,2
	mov physaddr,-1
	and ebx,0F0000000h
	mov ecx,PC
	or eax,ebx
	add ecx,4
	mov PC,eax
	mov r31,ecx
	ADD_CYCLES 2
	EXECUTE_DBSI
	RETURN

	
; BEQ rs,rt,imm16
opcode_04:
	COND_BRANCH je

; BNE rs,rt,imm16
opcode_05:

	COND_BRANCH jne

; BLEZ rs,imm16
opcode_06:
	COND_BRANCH_ZR jle

; BGTZ rs,imm16	
opcode_07:
	COND_BRANCH_ZR jg

; ADDI rt,rs,imm16
opcode_08:
	GET_ST
	and eax,0FFFFh
	mov ebx,[r0+ebx*4]
	add ebx,eax
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; ADDIU rt,rs,imm16
opcode_09:
	GET_ST
	and eax,0FFFFh
	mov ebx,[r0+ebx*4]
	add ebx,eax
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; SLTI rt,rs,imm16	
opcode_0A:
	GET_ST
	and eax,0FFFFh
	cmp eax,[r0+ebx*4]
	setl bl
	and ebx,1
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; SLTIU rt,rs,imm16
opcode_0B:
	GET_ST
	and eax,0FFFFh
	cmp eax,[r0+ebx*4]
	setb bl
	and ebx,1
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; ANDI rt,rs,imm16
opcode_0C:
	GET_ST
	and eax,0FFFFh
	mov ebx,[r0+ebx*4]
	and ebx,eax
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN
	
; ORI rt,rs,imm16
opcode_0D:
	GET_ST
	and eax,0FFFFh
	mov ebx,[r0+ebx*4]
	or ebx,eax
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN

; XORI rt,rs,imm16	
opcode_0E:
	GET_ST
	and eax,0FFFFh
	mov ebx,[r0+ebx*4]
	xor ebx,eax
	mov [r0+ecx*4],ebx
	ADD_CYCLES 1
	RETURN
	
; LUI rt,imm16
opcode_0F:
	GET_T
	and eax,0FFFFh
	and dword ptr [r0+ecx*4],0FFFFh
	shl eax,16
	or [r0+ecx*4],eax
	ADD_CYCLES 1
	RETURN
	
opcode_10:

; Coprocessor 1 instruction
opcode_11:
	mov ebx,eax
	and ebx,03E00000h
	.if ebx == 02000000h	; Single precision
		GET_COP_DST
		and eax,03Fh
		; ADD.S rd,rs,rt
		.if eax == 000000h
			fld dword ptr [f0+ebx*4]
			fadd dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; SUB.S rd,rs,rt
		.elseif eax == 000001h
			fld dword ptr [f0+ebx*4]
			fsub dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; MUL.S rd,rs,rt
		.elseif eax == 000002h
			fld dword ptr [f0+ebx*4]
			fmul dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 4
		; DIV.S rd,rs,rt
		.elseif eax == 000003h
			fld dword ptr [f0+ebx*4]
			fdiv dword ptr [f0+ecx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 40
		; ABS.S rd,rs
		.elseif eax == 000005h
			fld dword ptr [f0+ebx*4]
			fabs
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; NEG.S rd,rs
		.elseif eax == 000007h
			fld dword ptr [f0+ebx*4]
			fchs
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 2
		; CVT.W.S rd,rs
		.elseif eax == 000024h
			fld dword ptr [f0+ebx*4]
			fistp dword ptr [f0+edx*4]
			ADD_CYCLES 1
		.endif
	.elseif ebx == 0
		GET_COP_ST
		and eax,03Fh
		; MFC1 rt,rs
		.if eax == 000000h
			mov eax,[f0+ebx*4]
			mov [r0+ecx*4],eax
			ADD_CYCLES 1
		.endif
	.elseif ebx == 800000h
		GET_COP_ST
		and eax,03Fh
		; MTC1 rt,rs
		.if eax == 0
			mov eax,[r0+ebx*4]
			mov [f0+ecx*4],eax
			ADD_CYCLES 1
		.endif
	.elseif ebx == 2800000h
		GET_COP_DST
		and eax,03Fh
		; CVT.S.W rd,rs,rt
		.if eax == 20h
			fild dword ptr [f0+ebx*4]
			fstp dword ptr [f0+edx*4]
			ADD_CYCLES 1
		.endif
	.endif
	RETURN
	
opcode_12:
opcode_13:
opcode_14:
opcode_15:
opcode_16:
opcode_17:
opcode_18:
opcode_19:
opcode_1A:
opcode_1B:
opcode_1C:
opcode_1D:
opcode_1E:
opcode_1F:

; LB rt,imm(rs)
opcode_20:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_byte,eax
	pop ebp
	pop ecx
	movsx eax,al
	mov [r0+ecx*4],eax
	ADD_CYCLES 2
	RETURN

; LH rt,imm(rs)	
opcode_21:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_word,eax
	pop ebp
	pop ecx
	movsx eax,ax
	mov [r0+ecx*4],eax
	ADD_CYCLES 2
	RETURN
	
opcode_22:

; LW rt,imm(rs)
opcode_23:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_dword,eax
	pop ebp
	pop ecx
	mov [r0+ecx*4],eax
	ADD_CYCLES 2
	RETURN
	
; LBU rt,imm(rs)
opcode_24:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_byte,eax
	pop ebp
	pop ecx
	;and eax,0FFh
	mov byte ptr [r0+ecx*4],al
	ADD_CYCLES 2
	RETURN

; LHU rt,imm(rs)	
opcode_25:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_word,eax
	pop ebp
	pop ecx
	;and eax,0FFFFh
	mov word ptr [r0+ecx*4],ax
	ADD_CYCLES 2
	RETURN
	
opcode_26:
opcode_27:

; SB rt,imm(rs)
opcode_28:
	GET_ST
	and eax,0FFFFh
	mov ecx,[r0+ecx*4]
	add eax,[r0+ebx*4]
	push ebp	
	invoke mmu_write_byte,eax,ecx
	ADD_CYCLES 2
	pop ebp
	RETURN

; SH rt,imm(rs)	
opcode_29:
	GET_ST
	and eax,0FFFFh
	mov ecx,[r0+ecx*4]
	add eax,[r0+ebx*4]
	push ebp
	;	pusha
	;	invoke wsprintf,ADDR szBuffer,ADDR szFormat3,ecx,eax
	;	invoke lstrlen,ADDR szBuffer
	;	invoke WriteConsole,hout,ADDR szBuffer,eax,ADDR dummy,NULL
	;	popa
		
	invoke mmu_write_word,eax,ecx
	ADD_CYCLES 2
	pop ebp
	RETURN


opcode_2A:

; SW rt,imm(rs)
opcode_2B:
	GET_ST
	and eax,0FFFFh
	mov ecx,[r0+ecx*4]
	add eax,[r0+ebx*4]
	push ebp
	invoke mmu_write_dword,eax,ecx
	ADD_CYCLES 2
	pop ebp
	RETURN

opcode_2C:
opcode_2D:
opcode_2E:
opcode_2F:

opcode_30:

; L.S
opcode_31:
	GET_ST
	and eax,0FFFFh
	add eax,[r0+ebx*4]
	push ecx
	push ebp
	invoke mmu_read_dword,eax
	pop ebp
	pop ecx
	mov [f0+ecx*4],eax
	ADD_CYCLES 2
	RETURN	

opcode_32:
opcode_33:
opcode_34:
opcode_35:
opcode_36:
opcode_37:
opcode_38:
opcode_39:
opcode_3A:
opcode_3B:
opcode_3C:
opcode_3D:
opcode_3E:
opcode_3F:

	
opcode:
	dd OFFSET opcode_00
	dd OFFSET opcode_01
	dd OFFSET opcode_02
	dd OFFSET opcode_03
	dd OFFSET opcode_04
	dd OFFSET opcode_05
	dd OFFSET opcode_06
	dd OFFSET opcode_07
	dd OFFSET opcode_08
	dd OFFSET opcode_09
	dd OFFSET opcode_0A
	dd OFFSET opcode_0B
	dd OFFSET opcode_0C
	dd OFFSET opcode_0D
	dd OFFSET opcode_0E
	dd OFFSET opcode_0F
	dd OFFSET opcode_10
	dd OFFSET opcode_11
	dd OFFSET opcode_12
	dd OFFSET opcode_13
	dd OFFSET opcode_14
	dd OFFSET opcode_15
	dd OFFSET opcode_16
	dd OFFSET opcode_17
	dd OFFSET opcode_18
	dd OFFSET opcode_19
	dd OFFSET opcode_1A
	dd OFFSET opcode_1B
	dd OFFSET opcode_1C
	dd OFFSET opcode_1D
	dd OFFSET opcode_1E
	dd OFFSET opcode_1F
	dd OFFSET opcode_20
	dd OFFSET opcode_21
	dd OFFSET opcode_22
	dd OFFSET opcode_23
	dd OFFSET opcode_24
	dd OFFSET opcode_25
	dd OFFSET opcode_26
	dd OFFSET opcode_27
	dd OFFSET opcode_28
	dd OFFSET opcode_29
	dd OFFSET opcode_2A
	dd OFFSET opcode_2B
	dd OFFSET opcode_2C
	dd OFFSET opcode_2D
	dd OFFSET opcode_2E
	dd OFFSET opcode_2F
	dd OFFSET opcode_30
	dd OFFSET opcode_31
	dd OFFSET opcode_32
	dd OFFSET opcode_33
	dd OFFSET opcode_34
	dd OFFSET opcode_35
	dd OFFSET opcode_36
	dd OFFSET opcode_37
	dd OFFSET opcode_38
	dd OFFSET opcode_39
	dd OFFSET opcode_3A
	dd OFFSET opcode_3B
	dd OFFSET opcode_3C
	dd OFFSET opcode_3D
	dd OFFSET opcode_3E
	dd OFFSET opcode_3F
	




cpu_init PROC commlink:DWORD	
	mov eax,commlink
	mov zombie,eax

	invoke GetStdHandle,STD_OUTPUT_HANDLE
	mov hout,eax

	pusha	
	mov edi,offset r0
	mov ecx,32
	mov eax,0
	rep stosd
	popa
	
	ret
cpu_init ENDP



cpu_reset PROC
	pusha

	invoke zombie,ASK,MMU_READ_BYTE_CALLBACK,0,0
	mov mmu_read_byte,eax
	invoke zombie,ASK,MMU_READ_WORD_CALLBACK,0,0
	mov mmu_read_word,eax
	invoke zombie,ASK,MMU_READ_DWORD_CALLBACK,0,0
	mov mmu_read_dword,eax

	invoke zombie,ASK,MMU_WRITE_BYTE_CALLBACK,0,0
	mov mmu_write_byte,eax
	invoke zombie,ASK,MMU_WRITE_WORD_CALLBACK,0,0
	mov mmu_write_word,eax
	invoke zombie,ASK,MMU_WRITE_DWORD_CALLBACK,0,0
	mov mmu_write_dword,eax

	invoke zombie,ASK,MMU_TRANSLATE_ADDRESS_CALLBACK,0,0
	mov mmu_translate_address,eax

	mov edi,offset r0
	mov ecx,32
	mov eax,0
	rep stosd

	popa
	ret
cpu_reset ENDP


cpu_close PROC
	ret
cpu_close ENDP


cpu_config PROC hWnd:DWORD
	ret
cpu_config ENDP


cpu_set_io_handler PROC mode:DWORD,port:DWORD,pfhandler:DWORD
	push ebx

	mov eax,port
	mov ebx,pfhandler
	
	.if mode==INPUT	
		mov [inpb_handlers + eax*4],ebx
	.else
		mov [outpb_handlers + eax*4],ebx
	.endif
	
	pop ebx
	ret
cpu_set_io_handler ENDP


cpu_get PROC what:DWORD
	.if what == CPU_CYCLE_COUNT
		mov eax,cycle
	.elseif what == CPU_CYCLE_SCALING
		mov eax,200h
	.elseif what == CPU_SET_IO_CALLBACK
		mov eax,offset cpu_set_io_handler
	.elseif what == CPU_IRQ_CALLBACK
		mov eax,offset cpu_irq
	.elseif what == CPU_INPB_CALLBACK
		mov eax,offset cpu_inpb
	.elseif what == CPU_OUTPB_CALLBACK
		mov eax,offset cpu_outpb
	.elseif what == CPU_SPEED
		mov eax,speed
	.elseif what == CPU_REG_PC
		mov eax,PC
	;.elseif what == CPU_REG
	;	mov eax,param1
	;	and eax,0Fh
	;	mov eax,[r0+eax*4]
	.elseif what == PLUGIN_NAME
		mov eax,offset cpuName
	.endif
	ret
cpu_get ENDP


cpu_set PROC what:DWORD,param1:DWORD,param2:DWORD
	.if what == CPU_REG_PC
		mov eax,param1
		mov PC,eax
	.elseif what == CPU_REG_SP
		mov eax,param1
		mov r29,eax
	.elseif what == CPU_CYCLE_COUNT
		mov eax,param1
		mov cycle,eax
	.elseif what == CPU_BREAKPOINT
		mov eax,param1
		mov breakpoint,eax
	.elseif what == CPU_SPEED
		mov eax,param1
		mov speed,eax
	.elseif what == CPU_IO_HANDLER
		mov eax,param1
		and param1,0FFh
		shr eax,8
		invoke cpu_set_io_handler,eax,param1,param2
	.endif
	ret
cpu_set ENDP



cpu_outpb PROC port:DWORD,dat:DWORD
	push ecx
	push dword ptr dat
	mov eax,port
	call dword ptr [outpb_handlers + eax*4]
	pop ecx
	ret
cpu_outpb ENDP


cpu_inpb PROC port:DWORD
	push ecx
	mov eax,port
	call dword ptr [inpb_handlers + eax*4]
	pop ecx
	ret
cpu_inpb ENDP


cpu_irq PROC address:DWORD
	ret
cpu_irq ENDP


cpu_execute PROC
	ret
cpu_execute ENDP


PUBLIC cpu_execute_until
cpu_execute_until: ; PROC threshold:DWORD
;	LOCAL lastCyc:DWORD
	
	pusha
	
	mov ecx,cycle
	mov physaddr,-1
	
	@@exe_loop:
		cmp ecx,[esp+36] ;threshold
		jge @@exe_done
		mov lastCyc,ecx

		;cmp physaddr,-1
		;jne @@physaddr_valid
		.if physaddr == -1
			mov eax,PC
			;invoke mmu_read_dword,eax
			invoke mmu_translate_address,eax
			add eax,4
			mov physaddr,eax
			mov ecx,[eax]
			mov eax,[eax-4]
			mov nextInstr,ecx
		.else
			mov ebx,physaddr
			mov eax,nextInstr
			add ebx,4
			mov ecx,[ebx]
			mov physaddr,ebx
			mov nextInstr,ecx
		.endif
			
		;@@physaddr_valid:
		;mov ebx,physaddr
		;mov eax,[ebx]
		;mov ecx,[ebx+4]
		;mov nextInstr,ecx
		
		;add physaddr,4
		add PC,4
		mov ebx,PC
		mov lastPC,ebx
	
		test eax,0FC000000h
		jnz @@not_rtype
			mov ebx,eax
			and ebx,03Fh
			;call dword ptr [rtype_op + ebx*4]
			;jmp @@next_instr
			mov ebp,offset @@next_instr
			jmp dword ptr [rtype_op + ebx*4]
		@@not_rtype:
			mov ebx,eax
			shr ebx,26
			and ebx,03Fh
			;call dword ptr [opcode + ebx*4]
			mov ebp,offset @@next_instr
			jmp dword ptr [opcode + ebx*4]
		@@next_instr:
		
		mov ecx,cycle
		
		; !! Temp fix !!
		cmp ecx,lastCyc
		jne @@exe_loop
		mov ecx,1
	 
	@@exe_done:
	;mov cycle,ecx
	popa
	
	mov eax,cycle
	ret 4
;cpu_execute_until ENDP


END LibMain
