! SH-2 fire effect for Zombie
! /Mic, 2004
!
! Compile with GNUSH


.TEXT
.ALIGN 2


.GLOBAL _start

_start:

	! Copy initialised data from ROM to RAM
	mov.l 	rom_data,r1
	mov.l 	ram_data,r2
	mov 	#0,r0
	
	copy_data:
		mov.l 	@r1+,r5
		mov.l 	r5,@(r0,r2)
		add 	#4,r0
		cmp/eq 	#132,r0
		bf 	copy_data
		
	mov 	#0,r0

	mov.l 	gpu_reg0,r7
	mov.b 	r0,@r7				! Disable screen


	! Setup palette
	mov 	#0,r8
	mov 	#0x1F,r5
	mov.l 	color_mask,r6
	mov 	#64,r0
	mov.l 	palette_address,r4
	
	setup_palette:
		mov 	r8,r2
		shll8 	r2
		shll2 	r2
		shll 	r2
		swap.b	r2,r2			! Palette data is little-endian
		mov.w 	r2,@r4			! @r4 = r8 << 11
		
		mov 	r8,r3
		mov 	r5,r2
		shll2 	r3
		shll8 	r2
		shll2 	r3
		shll2 	r2
		shll2 	r3
		shll	r2
		or 	r3,r2
		swap.b	r2,r2
		mov.w 	r2,@(r0,r4)		! @(64,r4) = 0x1F<<11 | r8<<6
		
		mov 	r6,r2
		shll 	r0
		or 	r8,r2
		swap.b	r2,r2
		mov.w 	r2,@(r0,r4)		! @(128,r4) = 0xFFE0 | r8
		shlr 	r0
		
		add 	#2,r4
		add 	#1,r8
		mov 	#0x20,r9
		cmp/eq 	r9,r8
		bf 	setup_palette
		
		
	! Set the remaining colors to white
	mov.l	palette_address,r4
	add 	#192,r4
	mov	#1,r8
	neg	r8,r8				! 0xFFFF = full white
	mov	#0,r0

	set_colors_to_white:
		mov.w	r8,@r4
		add 	#2,r4
		add	#1,r0
		cmp/eq	#160,r0
		bf set_colors_to_white

	mov	#0xB4,r8
	mov.b	r8,@r7				! Enable display, vblank interrupts, bitmapped mode


	forever:
		! Wait for vertical blank
		mov.l	gpu_reg1,r5
		mov	#0x40,r9
		shll	r9
		wait_vb:
			mov.b	@r5,r0 
			and	r9,r0		! Check vblank bit
			cmp/eq #0x80,r0
			bf wait_vb
			
	
		mov #0,r8			! X
		x_loop:
			mov.l	video_memory,r4
			add	r8,r4
			mov.l	read_offset,r5
			mov.l	write_offset,r6
			add	r4,r5
			add	r4,r6
			mov	#71,r9		! Start at line 71
			y_loop:
				mov.l 	two_lines,r7
				add	#4,r7
				add	r5,r7
				
				! Read 4 pixels
				mov.b	@r5,r10
				mov.b	@(2,r5),r0
				mov	r0,r11
				mov.b	@(4,r5),r0
				mov	r0,r12
				mov.b	@r7,r13

				! Clear upper 24 bits of r10..r13
				mov	#255,r14
				extu.b	r14,r14
				and	r14,r10
				and	r14,r11
				and	r14,r12
				and	r14,r13
				
				! Calculate average
				add	r11,r10
				add	r12,r10
				add	r13,r10
				mov 	r10,r0
				shlr2	r0
				
				!Decrement color value if >0
				cmp/eq 	#0,r0
				bt 	no_dec
				dt 	r0
				no_dec:

				! Duplicate bits 0..7 into bits 8..15
				mov	r0,r10
				shll8	r10
				or	r10,r0
				
				mov.l	one_line,r4
				add	r6,r4
				mov.w	r0,@r6	! Draw 2 pixels
				mov.w	r0,@r4	! Draw 2 more pixels

				mov.l	two_lines,r10
				add	r10,r5	! Move 2 rows down
				add	r10,r6	! Dito
				
				mov	#239,r11
				extu.b	r11,r11
				add	#2,r9
				cmp/eq	r11,r9
				bf	y_loop

			bsr	Random
			nop

			add	#28,r2
			mov 	#191,r10
			extu.b	r10,r10
			and	r10,r2
			
			mov.l	video_memory,r4
			mov.l	screen_offset,r11
			add	r8,r4		! Add X
			add	r11,r4
			mov.b	r2,@r4		! Store random pixel

			add	#2,r8
			mov.l	one_line,r10
			cmp/eq	r10,r8
			bf	x_loop
	
		bra forever			


	Random:
		mov.l	ram_data,r10		! Point to seed1 in RAM
		mov	r10,r3
		add	#2,r3			! Point to the random number table in RAM
		
		mov.b	@r10,r11		! r11 = seed1
		mov.b	@(1,r10),r0		! r0 = seed2
		mov	#0xFF,r14
		extu.b	r14,r14
		and	r14,r11			! Clear upper 24 bits
		and	r14,r0
		mov	r11,r12
		mov	r0,r13
		shll	r12
		shll	r13
		add	r3,r12
		add	r3,r13
		mov.w	@r12,r12		! Load random number
		mov.w	@r13,r14		! ..another one
		add	r14,r12			! Add the two numbers
		mov.w	r12,@r13		! Store it
		add	#1,r11			! Increase seed1
		add	#1,r0			! Increase seed2
		mov	#0x3F,r14		! The table contains 64 entries
		and	r14,r11
		and	r14,r0
		mov	r12,r2			! Put the random number in r2
		mov.b	r11,@r10		! Store seed1
		add	#1,r10
		mov.b	r0,@r10			! Store seed2

		rts
		

.align 4

video_memory:
	.LONG 0x6000000
palette_address:
	.LONG 0x6012C00
color_mask:
	.LONG 0xFFE0
gpu_reg0:
	.LONG 0xFF00
gpu_reg1:
	.LONG 0xFF01
read_offset:
	.LONG 22718
write_offset:
	.LONG 22080
two_lines:
	.LONG 640
one_line:
	.LONG 320
screen_offset:
	.LONG 75840
last_y:
	.LONG 239
	
rom_data:
	.LONG rng_data_rom
ram_data:
	.LONG data_ram
	
rng_data_rom:	
	! Seeds
	.BYTE 29
	.BYTE 63

	! 64 random 16-bit numbers for use with the random-number generator 
	.WORD 25662,56977,36314,46707
	.WORD 56747,47936,28467,50155
	.WORD 54470,9126,19191,26274
	.WORD 27205,13469,6769,48786
	.WORD 43195,17969,58549,13097
	.WORD 9962,26613,60724,25428
	.WORD 44416,43884,31777,40699
	.WORD 4280,32767,31882,41912
	.WORD 19050,42386,38267,39411
	.WORD 11739,36780,10921,43911
	.WORD 7473,38901,28614,62047
	.WORD 25094,22064,16149,14666
	.WORD 46209,34643,30017,25842
	.WORD 64242,3527,34102,35358
	.WORD 2277,3678,16479,3431
	.WORD 36135,63741,11311,0
          

.DATA
data_ram:
	seed1: .BYTE
	seed2: .BYTE
	random_tbl:
	